print("import: 'pyomo'")
import pyomo

print("import: 'pyomo.bilevel'")
import pyomo.bilevel

print("import: 'pyomo.bilevel.plugins'")
import pyomo.bilevel.plugins

print("import: 'pyomo.checker'")
import pyomo.checker

print("import: 'pyomo.checker.plugins'")
import pyomo.checker.plugins

print("import: 'pyomo.checker.plugins.checkers'")
import pyomo.checker.plugins.checkers

print("import: 'pyomo.checker.plugins.checkers.model'")
import pyomo.checker.plugins.checkers.model

print("import: 'pyomo.checker.plugins.checkers.py3k'")
import pyomo.checker.plugins.checkers.py3k

print("import: 'pyomo.checker.plugins.checkers.sample'")
import pyomo.checker.plugins.checkers.sample

print("import: 'pyomo.common'")
import pyomo.common

print("import: 'pyomo.common.collections'")
import pyomo.common.collections

print("import: 'pyomo.contrib'")
import pyomo.contrib

print("import: 'pyomo.contrib.appsi'")
import pyomo.contrib.appsi

print("import: 'pyomo.contrib.appsi.cmodel'")
import pyomo.contrib.appsi.cmodel

print("import: 'pyomo.contrib.appsi.examples'")
import pyomo.contrib.appsi.examples

print("import: 'pyomo.contrib.appsi.solvers'")
import pyomo.contrib.appsi.solvers

print("import: 'pyomo.contrib.appsi.utils'")
import pyomo.contrib.appsi.utils

print("import: 'pyomo.contrib.appsi.writers'")
import pyomo.contrib.appsi.writers

print("import: 'pyomo.contrib.benders'")
import pyomo.contrib.benders

print("import: 'pyomo.contrib.community_detection'")
import pyomo.contrib.community_detection

print("import: 'pyomo.contrib.example'")
import pyomo.contrib.example

print("import: 'pyomo.contrib.example.plugins'")
import pyomo.contrib.example.plugins

print("import: 'pyomo.contrib.fbbt'")
import pyomo.contrib.fbbt

print("import: 'pyomo.contrib.fme'")
import pyomo.contrib.fme

print("import: 'pyomo.contrib.gdp_bounds'")
import pyomo.contrib.gdp_bounds

print("import: 'pyomo.contrib.gdpbb'")
import pyomo.contrib.gdpbb

print("import: 'pyomo.contrib.gdpopt'")
import pyomo.contrib.gdpopt

print("import: 'pyomo.contrib.incidence_analysis'")
import pyomo.contrib.incidence_analysis

print("import: 'pyomo.contrib.incidence_analysis.common'")
import pyomo.contrib.incidence_analysis.common

print("import: 'pyomo.contrib.interior_point'")
import pyomo.contrib.interior_point

print("import: 'pyomo.contrib.interior_point.linalg'")
import pyomo.contrib.interior_point.linalg

print("import: 'pyomo.contrib.mcpp'")
import pyomo.contrib.mcpp

print("import: 'pyomo.contrib.mindtpy'")
import pyomo.contrib.mindtpy

print("import: 'pyomo.contrib.multistart'")
import pyomo.contrib.multistart

print("import: 'pyomo.contrib.parmest'")
import pyomo.contrib.parmest

print("import: 'pyomo.contrib.parmest.examples'")
import pyomo.contrib.parmest.examples

print("import: 'pyomo.contrib.parmest.examples.reactor_design'")
import pyomo.contrib.parmest.examples.reactor_design

print("import: 'pyomo.contrib.parmest.examples.rooney_biegler'")
import pyomo.contrib.parmest.examples.rooney_biegler

print("import: 'pyomo.contrib.parmest.examples.semibatch'")
import pyomo.contrib.parmest.examples.semibatch

print("import: 'pyomo.contrib.preprocessing'")
import pyomo.contrib.preprocessing

print("import: 'pyomo.contrib.preprocessing.plugins'")
import pyomo.contrib.preprocessing.plugins

print("import: 'pyomo.contrib.pynumero'")
import pyomo.contrib.pynumero

print("import: 'pyomo.contrib.pynumero.algorithms'")
import pyomo.contrib.pynumero.algorithms

print("import: 'pyomo.contrib.pynumero.algorithms.solvers'")
import pyomo.contrib.pynumero.algorithms.solvers

print("import: 'pyomo.contrib.pynumero.examples'")
import pyomo.contrib.pynumero.examples

print("import: 'pyomo.contrib.pynumero.examples.callback'")
import pyomo.contrib.pynumero.examples.callback

print("import: 'pyomo.contrib.pynumero.examples.external_grey_box'")
import pyomo.contrib.pynumero.examples.external_grey_box

print("import: 'pyomo.contrib.pynumero.examples.external_grey_box.param_est'")
import pyomo.contrib.pynumero.examples.external_grey_box.param_est

print("import: 'pyomo.contrib.pynumero.examples.external_grey_box.react_example'")
import pyomo.contrib.pynumero.examples.external_grey_box.react_example

print("import: 'pyomo.contrib.pynumero.interfaces'")
import pyomo.contrib.pynumero.interfaces

print("import: 'pyomo.contrib.pynumero.linalg'")
import pyomo.contrib.pynumero.linalg

print("import: 'pyomo.contrib.pynumero.sparse'")
import pyomo.contrib.pynumero.sparse

print("import: 'pyomo.contrib.pyros'")
import pyomo.contrib.pyros

print("import: 'pyomo.contrib.satsolver'")
import pyomo.contrib.satsolver

print("import: 'pyomo.contrib.sensitivity_toolbox'")
import pyomo.contrib.sensitivity_toolbox

print("import: 'pyomo.contrib.sensitivity_toolbox.examples'")
import pyomo.contrib.sensitivity_toolbox.examples

print("import: 'pyomo.contrib.trustregion'")
import pyomo.contrib.trustregion

print("import: 'pyomo.contrib.viewer'")
import pyomo.contrib.viewer

print("import: 'pyomo.core'")
import pyomo.core

print("import: 'pyomo.core.base'")
import pyomo.core.base

print("import: 'pyomo.core.beta'")
import pyomo.core.beta

print("import: 'pyomo.core.expr'")
import pyomo.core.expr

print("import: 'pyomo.core.expr.calculus'")
import pyomo.core.expr.calculus

print("import: 'pyomo.core.kernel'")
import pyomo.core.kernel

print("import: 'pyomo.core.kernel.piecewise_library'")
import pyomo.core.kernel.piecewise_library

print("import: 'pyomo.core.plugins'")
import pyomo.core.plugins

print("import: 'pyomo.core.plugins.transform'")
import pyomo.core.plugins.transform

print("import: 'pyomo.dae'")
import pyomo.dae

print("import: 'pyomo.dae.plugins'")
import pyomo.dae.plugins

print("import: 'pyomo.dataportal'")
import pyomo.dataportal

print("import: 'pyomo.dataportal.plugins'")
import pyomo.dataportal.plugins

print("import: 'pyomo.duality'")
import pyomo.duality

print("import: 'pyomo.environ'")
import pyomo.environ

print("import: 'pyomo.gdp'")
import pyomo.gdp

print("import: 'pyomo.gdp.plugins'")
import pyomo.gdp.plugins

print("import: 'pyomo.kernel'")
import pyomo.kernel

print("import: 'pyomo.mpec'")
import pyomo.mpec

print("import: 'pyomo.mpec.plugins'")
import pyomo.mpec.plugins

print("import: 'pyomo.neos'")
import pyomo.neos

print("import: 'pyomo.neos.plugins'")
import pyomo.neos.plugins

print("import: 'pyomo.network'")
import pyomo.network

print("import: 'pyomo.network.plugins'")
import pyomo.network.plugins

print("import: 'pyomo.opt'")
import pyomo.opt

print("import: 'pyomo.opt.base'")
import pyomo.opt.base

print("import: 'pyomo.opt.parallel'")
import pyomo.opt.parallel

print("import: 'pyomo.opt.plugins'")
import pyomo.opt.plugins

print("import: 'pyomo.opt.problem'")
import pyomo.opt.problem

print("import: 'pyomo.opt.results'")
import pyomo.opt.results

print("import: 'pyomo.opt.solver'")
import pyomo.opt.solver

print("import: 'pyomo.opt.testing'")
import pyomo.opt.testing

print("import: 'pyomo.repn'")
import pyomo.repn

print("import: 'pyomo.repn.beta'")
import pyomo.repn.beta

print("import: 'pyomo.repn.plugins'")
import pyomo.repn.plugins

print("import: 'pyomo.repn.plugins.ampl'")
import pyomo.repn.plugins.ampl

print("import: 'pyomo.scripting'")
import pyomo.scripting

print("import: 'pyomo.scripting.plugins'")
import pyomo.scripting.plugins

print("import: 'pyomo.solvers'")
import pyomo.solvers

print("import: 'pyomo.solvers.plugins'")
import pyomo.solvers.plugins

print("import: 'pyomo.solvers.plugins.converter'")
import pyomo.solvers.plugins.converter

print("import: 'pyomo.solvers.plugins.solvers'")
import pyomo.solvers.plugins.solvers

print("import: 'pyomo.util'")
import pyomo.util

print("import: 'pyomo.version'")
import pyomo.version

